-- AVI file structures
-- (c) FSC, 2007

--
-- auxiliary functions
--
function QuoteString(data)
    return '"' .. data.Data .. '"'
end

function GetQuotedChunkName()
    return QuoteString(PeekBytes(4))
end

function CheckChunkID(chunk, err)
    if PeekBytes(4).Data ~= chunk then error(err) end
end

--
-- global variables to count different chunks
--
NumList = 0
NumJunk = 0
NumUnknown = 0

--
-- ChunkID type that is basically a DWORD but is shown as DWORD ID = "LIST"
-- This is not entirely correct in C sense (a better definition would be
-- char ID[4] = { 'L', 'I', 'S', 'T' } or DWORD ID = 'TSIL' but such presentations are less obvious
-- so we'll stick with this one
--
ChunkID = type_ui32("DWORD", function() Str() return "ID" end)

--
-- MainAVIHeader.dwFlags
--
local AVIF =
{
    AVIF_HASINDEX       =   0x00000010,
    AVIF_MUSTUSEINDEX   =   0x00000020,
    AVIF_ISINTERLEAVED  =   0x00000100,
    AVIF_TRUSTCKTYPE    =   0x00000800,
    AVIF_WASCAPTUREFILE =   0x00010000,
    AVIF_COPYRIGHTED    =   0x00020000
}

local function MainAVIHeaderFlags()
    StartType("DWORD", "dwFlags")
    local Val = GetUI32()
    PrintNumber(BitSetFmt(Val, AVIF, "0x%x"))
    EndType(string.format("0x%x", Val))
end

--
-- MainAVIHeader
--
local function _MainAVIHeader()
    TypeWidth = 6
    NameWidth = 22
    
    Dec()
    local FPS = PeekUI32()
    local FPSComment = ""
    if FPS ~= 0 then FPSComment = "(" .. 1e6 / FPS .. " frames/second) " end
    DWORD("dwMicroSecPerFrame", FPSComment .. "frame display rate (or 0L)")
    DWORD("dwMaxBytesPerSec", "max. transfer rate")
    
    if PeekUI32() == 0 then Dec() else Hex() end
    DWORD("dwPaddingGranularity", "pad to multiples of this size; normally 2K.")
    
    MainAVIHeaderFlags()
    
    Dec()
    DWORD("dwTotalFrames", "# frames in file")
    DWORD "dwInitialFrames"
    DWORD "dwStreams"
    
    if PeekUI32() == 0 then Dec() else HexDec() end
    DWORD "dwSuggestedBufferSize"

    Dec()
    DWORD "dwWidth"
    DWORD "dwHeight"

    DWORD("dwReserved", nil, 4)
end

MainAVIHeader = struct("MainAVIHeader", _MainAVIHeader)

local function _AVIH()
    TypeWidth = 6
    NameWidth = 6
    
    CheckChunkID("avih", 'Not an "avih" chunk')
    ChunkID()
    
    local Len = PeekUI32()
    
    HexDec()
    DWORD "Size"

    local StartPos = GetFilePosition()
    CrLf()
    MainAVIHeader("AVIHeader")
    
    local NumBytes = BytesProcessed(StartPos)
    if NumBytes < Len then BYTE("ExtraData", "???", Len - NumBytes)
    elseif NumBytes > Len then error('"hdrl": Size is smaller than sizeof(MainAVIHeader)') end
end

AVIH = struct("AVIH", _AVIH)

--
-- "hdrl" chunk
--
local function _HDRL()
    TypeWidth = 5
    NameWidth = 5
    
    CheckChunkID("hdrl", 'Not a "hdrl" chunk')
    ChunkID()    

    AVIH("avih")
end

HDRL = struct("HDRL", _HDRL)

--
-- RECT
--
local function _SMALL_RECT()
    TypeWidth = 6
    NameWidth = 7

    Dec()
    short "left"
    short "top"
    short "right"
    short "bottom"
end

SMALL_RECT = struct("SMALL_RECT", _SMALL_RECT)

--
-- "strh" chunk
--
local AVISF = { AVISF_DISABLED = 0x00000001, AVISF_VIDEO_PALCHANGES = 0x00010000 }

local function AVISHFlags()
    StartType("DWORD", "dwFlags")
    local Val = GetUI32()
    PrintNumber(BitSetFmt(Val, AVISF, "0x%x"))
    local Comment
    if Val ~= 0 then Comment = string.format("0x%x", Val) end
    EndType(Comment)
end

FOURCC = type_ui32("FOURCC", function() Str() end)

local function _AVIStreamHeader()
    TypeWidth = 7
    NameWidth = 22
    
    FOURCC "fccType"
    FOURCC "fccHandler"
    
    AVISHFlags()

    Dec()
    DWORD  "dwPriority"
    DWORD  "dwInitialFrames"
    DWORD  "dwScale"
    DWORD  "dwRate"
    DWORD  "dwStart"
    HexDec()
    DWORD  "dwLength"
    DWORD  "dwSuggestedBufferSize"
    Dec()
    DWORD  "dwQuality"
    DWORD  "dwSampleSize"
    
    NameWidth = 1
    SMALL_RECT "rcFrame"
end

AVIStreamHeader = struct("AVIStreamHeader", _AVIStreamHeader)

function _STRH()
    TypeWidth = 5
    NameWidth = 5
    
    CheckChunkID("strh", 'Not a "strh" chunk')
    ChunkID()
    
    local Len = PeekUI32()
    HexDec()
    DWORD "Size"    
    
    local StartPos = GetFilePosition()
    CrLf()
    AVIStreamHeader("StreamHeader")
    
    local NumBytes = BytesProcessed(StartPos)
    if NumBytes < Len then BYTE("ExtraData", "???", Len - NumBytes)
    elseif NumBytes > Len then error('"strh": Size is smaller than sizeof(AVIStreamHeader)') end
end

STRH = struct("STRH", _STRH)
--
-- "strl" chunk
--
function _STRL()
    TypeWidth = 5
    NameWidth = 5
    
    CheckChunkID("strl", 'Not a "strl" chunk')
    ChunkID()    
    
    STRH("strh")
end

STRL = struct("STRL", _STRL)

--
-- "LIST" chunk
--
function _LIST()
    NumList = NumList + 1

    CheckChunkID("LIST", "Not a 'LIST' chunk")
    ChunkID()    
    
    local Len = PeekUI32();
    HexDec()
    DWORD "Size"
    
    while Len > 0 do
        local StartPos = GetFilePosition()
        
        local ListType = PeekBytes(4).Data
        if ListType == "hdrl" then HDRL("hdrl")
        elseif ListType == "strl" then STRL("strl")
        elseif ListType == "LIST" then LIST("List" .. NumList)
        elseif ListType == "JUNK" then JUNK("Junk" .. NumJunk)
        else
            ChunkID()
    
            Hex()
            BYTE("Data", nil, Len - 4)
        end
        
        Len = Len - BytesProcessed(StartPos)
        if Len < 0 then error('Invalid Size in "LIST" chunk') end
    end
end

LIST = struct("LIST", _LIST)

--
-- "JUNK" chunk
--
function _JUNK()
    NumJunk = NumJunk + 1

    CheckChunkID("JUNK", "Not a 'JUNK' chunk")
    ChunkID()    
    
    local Len = PeekUI32();
    HexDec()
    DWORD "Size"
    
    Hex()
    BYTE("Data", nil, Len)
end

JUNK = struct("JUNK", _JUNK)
--
-- unknown chunk
--
function _UNKNOWNCHUNK()
    NumUnknown = NumUnknown + 1
    
    ChunkID()

    local Len = PeekUI32();
    HexDec()
    DWORD "Size"
    
    Hex()
    BYTE("Data", nil, Len)
end

UNKNOWNCHUNK = struct("UNKNOWN", _UNKNOWNCHUNK)

